import numpy as np
from candl.tensor import Tensor

class Parameter: 
  def __init__(self, data, requires_grad=True):
    self.data = Tensor(data, requires_grad, float) # Set data of parameter

class Module: 
  def __init__(self, *args): 
    self.__children = args or [] # Initialize module with given arguments as children
        
  def forward(self, x):
    # Call .forward() on all of the children
    for child in self.__children:
      x = child.forward(x)
    return x

  def parameters(self):
    # Fetch all of the parameters by going through each child and getting all of the Parameter instances
    parameters = []
    for child in self.__children:
      for (key, value) in child.__dict__.items():
        if isinstance(value, Parameter):
          parameters.append(value)
    return parameters
      
# Just to make it more pytorchy, allows you to chain together modules to make a "model" (or just an arbitrary chain of modules)
def Sequential(*args): 
  return Module(*args)

# Just a basic linear layer with an optional bias node
class Linear(Module):
  def __init__(self, input_neurons, output_neurons, bias=True):
    self.weights = Parameter(np.random.normal(0, (2 / input_neurons)**1/2, (input_neurons, output_neurons)))
    self.bias = bias
    if bias:
      self.biases = Parameter(np.zeros(output_neurons))

  def forward(self, x):
    out = x.matmul(self.weights.data)
    if self.bias: 
      return out + self.biases.data
    return out
  
class ReLU(Module):
  def forward(self, x):
    return x.relu()

class Sigmoid(Module):
  def forward(self, x):
    return x.sigmoid()

class Tanh(Module):
  def forward(self, x):
    return x.tanh()

class SGD: 
  def __init__(self, parameters, learning_rate):
    self.parameters = parameters
    self.learning_rate = learning_rate

  def step(self, zero_grad=False):
    for param in self.parameters: 
      param = param.data # Get the data inside the parameter
      # w_i+1 <- w_i - lr * (dL/dw_i)
      param.els = param.els - self.learning_rate * param.grad
      # Reset the gradient if specified
      if zero_grad:
        param.zero_grad()
    
  def zero_grad(self): 
    for param in self.parameters: 
      param.zero_grad()

def MSE(): 
  # Compute the mean squared error loss given a prediction (yhat) and a target (y)
  def compute_loss(yhat, y):    
    N = np.prod(y.els.shape, dtype=float)
    return ((y - yhat) ** 2).sum() / N
  return compute_loss


import numpy as np

class Tensor:
  def __init__(self, els, requires_grad=False, dtype=None, tape=None):
    # Store the elements of our tensor in a numpy ndarray
    self.els = np.array(els, dtype=dtype)
    # Decide whether or not we are computing gradients
    self.requires_grad = requires_grad
    # Initialize an empty gradient tape for autodiff
    self.tape = tape or []
    # If the dtype is not "float", raise an error
    if not self.els.dtype == float and requires_grad:
        raise TypeError("Autodiff is only allowed for tensors of type \"float\"!")
    # Otherwise, initialize our gradient
    if requires_grad:
        self.zero_grad()
    else: 
      self.grad = None

  def __getitem__(self, n):
    return self.els[n]

  def __repr__(self):
    return self.els.__str__()

  def __op(self, other, operation, derivative_a, derivative_b=None):
    # Turn the other argument into a tensor if it isn't already one
    if not isinstance(other, Tensor):
      other = Tensor([other], dtype=self.els.dtype, requires_grad=False)
    # If either one of our tensors require grad, then the output tensor also requires grad
    requires_grad = self.requires_grad or other.requires_grad
    # Find the output tensor by executing the given operation
    out = Tensor(operation(self.els, other.els), requires_grad=requires_grad, tape=(other.tape + self.tape))
    # Push a function into the tape that we can call during reverse mode
    def tape_action(): 
      # Execute the derivative functions for the operation if the gradients exist
      if self.grad is not None: 
        self.grad = self.grad + derivative_a(self.els, other.els, out.grad)
        if self.grad.shape != self.els.shape:
          self.grad = np.add.reduce(self.grad)
      if other.grad is not None and derivative_b is not None:
        other.grad = other.grad + derivative_b(self.els, other.els, out.grad)
        if other.grad.shape != other.els.shape:
          other.grad = np.add.reduce(other.grad)
    out.tape.append(tape_action)
    # Return the outputted value
    return out

  def __add__(self, other):
    return self.__op(other, lambda a, b: a + b, lambda a, b, dc: dc, lambda a, b, dc: dc)

  def __sub__(self, other):
    return self.__op(other, lambda a, b: a - b, lambda a, b, dc: dc, lambda a, b, dc: -dc)

  def __rsub__(self, other):
    return -self + other

  def __mul__(self, other):
    return self.__op(other, lambda a, b: a * b, lambda a, b, dc: b * dc, lambda a, b, dc: a * dc)

  def __pow__(self, other): 
    return self.__op(other, lambda a, b: a ** b, lambda a, b, dc: b * a ** (b - 1) * dc, lambda a, b, dc: np.log(a) * a ** b * dc)

  def __neg__(self): 
    return self * -1

  def __truediv__(self, other):
    return self * other ** -1

  def __rtruediv__(self, other):
    return self ** -1 * other

  def sum(self):
    return self.__op(1, lambda a, b: np.sum(a), lambda a, b, dc: dc)

  def relu(self): 
    return self.__op(1, lambda a, b: a * (a > 0), lambda a, b, dc: (a > 0) * dc)

  def sigmoid(self): 
    sigmoid = lambda a, b: 1 / (1 + np.exp(-a))
    return self.__op(1, sigmoid, lambda a, b, dc: sigmoid(a, b) * (1 - sigmoid(a, b)) * dc)

  def tanh(self): 
    tanh = lambda a, b: np.tanh(a)
    return self.__op(1, tanh, lambda a, b, dc: 1 - tanh(a, b) ** 2 * dc)

  def matmul(self, other):
    return self.__op(other, lambda a, b: np.dot(a, b), lambda a, b, dc: np.dot(dc, b.T), lambda a, b, dc: np.dot(a.T, dc)) 

  def backward(self): 
    self.grad = np.ones(self.els.shape) # Set the seed
    while len(self.tape) > 0: 
      self.tape.pop()() # Remove and execute the last element of the tape
    return self

  def zero_grad(self): 
    self.grad = np.zeros_like(self.els)

  __radd__ = __add__
  __rmul__ = __mul__